#############################################################################
#
#    APSYNSIM: A real-time Aperture Synthesis Simulator
#
#    Copyright (C) 2014  Ivan Marti-Vidal (Nordic ARC Node, OSO, Sweden)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#############################################################################


import Tkinter
import FileDialog
import pylab as pl
import numpy as np
import scipy.ndimage.interpolation as spndint
from matplotlib.widgets import Slider, Button
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.cm as cm
import matplotlib.image as plimg
import tkFileDialog
from tkMessageBox import showinfo
import os
import time

__version__ = '0.3.1'

class Interferometer(object):

  def __init__(self,antenna_file="",model_file=""):

    self.__version__ = __version__

    self.Hfac = np.pi/180.*15.
    self.deg2rad = np.pi/180.

    self.robust = 0.0
    self.deltaAng = 1.*self.deg2rad
    self.gamma = 0.5  # Gamma correction to plot model.
    self.lfac = 1.e6   # Lambda units (i.e., 1.e6 => Mlambda)
    self.ulab = r'U (M$\lambda$)'
    self.vlab = r'V (M$\lambda$)'


# Default of defaults!
    nH = 200
    Npix = 512   # Image pixel size. Must be a power of 2
    DefaultMod = 'Nebula.model'
    DefaultArray = 'Golay_12.array'

# Overwrite defaults from config file:
    d1 = os.path.dirname(os.path.realpath(__file__))
    print d1

#   execfile(os.path.join(os.path.basename(d1),'APSYNSIM.config'))
    conf = open(os.path.join(d1,'APSYNSIM.config'))
    for line in conf.readlines():
      temp=line.replace(' ','')
      if len(temp)>2:
         if temp[0:4] == 'Npix':
           Npix = int(temp[5:temp.find('#')])
         if temp[0:2] == 'nH':
           nH = int(temp[3:temp.find('#')])
         if temp[0:10] == 'DefaultMod':
           DefaultModel = temp[12:temp.find('#')].replace('\'','').replace('\"','')
         if temp[0:12] == 'DefaultArray':
           DefaultArray = temp[14:temp.find('#')].replace('\'','').replace('\"','')

    conf.close()

# Set instance configuration values:
    self.nH = nH
    self.Npix = Npix

    self.datadir = os.path.join(d1,'..','PICTURES')
    self.arraydir  = os.path.join(d1,'..','ARRAYS')
    self.modeldir  = os.path.join(d1,'..','SOURCE_MODELS')

 # Try to read a default initial array:
    if len(antenna_file)==0:
      try:
        antenna_file = os.path.join(self.arraydir,DefaultArray)
      except:
        pass

 # Try to read a default initial model:
    if len(model_file)==0:
      try:
        model_file = os.path.join(self.modeldir,DefaultModel)
      except:
        pass


    self.lock=False
    self._onSphere = False

    self.readAntennas(str(antenna_file))
    self.readModels(str(model_file))
    self.GUI()



  def showError(self,message):
    showinfo('ERROR!', message)
    raise Exception(message)



  def GUI(self):

    self.Nphf = self.Npix/2
    self.robfac = 0.0
    self.figUV = pl.figure(figsize=(15,8))
    self.figUV.canvas.set_window_title('Aperture Synthesis Simulator (I. Marti-Vidal, Onsala Space Observatory) - version  %s'%__version__)
    self.antPlot = self.figUV.add_subplot(231,aspect='equal')
    self.UVPlot = self.figUV.add_subplot(232,aspect='equal',axisbg=(0.4,0.4,0.4))
    self.beamPlot = self.figUV.add_subplot(233,aspect='equal')
    self.modelPlot = self.figUV.add_subplot(235,aspect='equal')
    self.dirtyPlot = self.figUV.add_subplot(236,aspect='equal')

    self.spherePlot = pl.axes([0.53,0.82,0.12,0.12],projection='3d',aspect='equal')

    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)
    x = 10 * np.outer(np.cos(u), np.sin(v))
    y = 10 * np.outer(np.sin(u), np.sin(v))
    z = 10 * np.outer(np.ones(np.size(u)), np.cos(v))
    self.spherePlotPlot = self.spherePlot.plot_surface(x, y, z,  rstride=4, cstride=4, color=(0.8,0.8,1.0))
    self.spherePlot._axis3don = False
    self.spherePlot.patch.set_alpha(0.8)
    beta = np.zeros(100)
    self.arrayPath = [np.zeros(self.nH), np.zeros(self.nH), np.zeros(self.nH)]
    self.sphereArray = self.spherePlot.plot([],[],[],'y',linewidth=3)
    self.spherePlot.set_xlim3d((-6,6))
    self.spherePlot.set_ylim3d((-6,6))
    self.spherePlot.set_zlim3d((-6,6))
    self.spherePlot.patch.set_alpha(0.8)
    self.spherePlot.elev = 45.


    self.figUV.subplots_adjust(left=0.05,right=0.99,top=0.95,bottom=0.07,hspace=0.25)
    self.figUV.canvas.mpl_connect('pick_event', self._onPick)
    self.figUV.canvas.mpl_connect('motion_notify_event', self._onAntennaDrag)
    self.figUV.canvas.mpl_connect('button_release_event',self._onRelease)
    self.figUV.canvas.mpl_connect('button_press_event',self._onPress)

    self.pickAnt = False

    self.fmtH = r'$\phi = $ %3.1f$^\circ$   $\delta = $ %3.1f$^\circ$' "\n" r'H = %3.1fh / %3.1fh'
    self.fmtBas = r'Bas %i $-$ %i  at  H = %4.2fh'
    self.fmtVis = r'Amp: %.1e Jy.   Phase: %5.1f deg.' 
    self.fmtA = 'N = %i'
    self.fmtA2 = '  Picked Ant. #%i'
    fmtB1 = r'$\lambda = $ %4.1fmm  '%(self.wavelength*1.e6)
    self.fmtB = fmtB1 + "\n" + r'% 4.2f Jy/beam' + "\n" + r'$\Delta\alpha = $ % 4.2f / $\Delta\delta = $ % 4.2f '
    self.fmtD = r'% .2e Jy/beam' "\n" r'$\Delta\alpha = $ % 4.2f / $\Delta\delta = $ % 4.2f '
    self.fmtM = r'%.2e Jy/pixel' "\n"  r'$\Delta\alpha = $ % 4.2f / $\Delta\delta = $ % 4.2f'

    self.wax = {}
    self.widget = {}
    self.wax['lat'] = pl.axes([0.07,0.45,0.25,0.04])
    self.wax['dec'] = pl.axes([0.07,0.40,0.25,0.04])
    self.wax['H0'] = pl.axes([0.07,0.35,0.25,0.04])
    self.wax['H1'] = pl.axes([0.07,0.30,0.25,0.04])
    self.wax['wave'] = pl.axes([0.07,0.25,0.25,0.04])
    self.wax['robust'] = pl.axes([0.07,0.20,0.25,0.04])
    self.wax['add'] = pl.axes([0.07,0.11,0.12,0.07])
    self.wax['rem'] = pl.axes([0.20,0.11,0.12,0.07])
    self.wax['save'] =  pl.axes([0.07,0.05,0.08,0.05]) 
    self.wax['loadarr']=pl.axes([0.155,0.05,0.08,0.05])
    self.wax['loadmod']=pl.axes([0.24,0.05,0.08,0.05])
    self.wax['gammacorr']=pl.axes([0.46,0.08,0.13,0.02],axisbg='white')
   # self.wax['gammacorr'].yaxis.label.set_color('white')
   # self.wax['gammacorr'].label.set_color('white')

    self.widget['robust'] = Slider(self.wax['robust'],r'Robust',-2.,2.,valinit=0.0)
    self.widget['lat'] = Slider(self.wax['lat'],r'Lat (deg)',-90.,90.,valinit=self.lat/self.deg2rad)
    self.widget['dec'] = Slider(self.wax['dec'],r'Dec (deg)',-90.,90.,valinit=self.dec/self.deg2rad)
    self.widget['H0'] = Slider(self.wax['H0'],r'H$_{0}$ (h)',-12.,12.,valinit=self.Hcov[0]/self.Hfac)
    self.widget['H1'] = Slider(self.wax['H1'],r'H$_{1}$ (h)',-12.,12.,valinit=self.Hcov[1]/self.Hfac)
    self.widget['wave'] = Slider(self.wax['wave'],r'$\lambda$ (mm)',self.wavelength*1.e6/10.,10.*self.wavelength*1.e6,valinit=self.wavelength*1.e6)
    self.widget['add'] = Button(self.wax['add'],r'+ Antenna')
    self.widget['rem'] = Button(self.wax['rem'],r'$-$ Antenna')
    self.widget['save'] = Button(self.wax['save'],'Save array')
    self.widget['loadarr'] = Button(self.wax['loadarr'],'Load array')
    self.widget['loadmod'] = Button(self.wax['loadmod'],'Load model')
    self.widget['gammacorr'] = Slider(self.wax['gammacorr'],'gamma',0.1,1.0,valinit=self.gamma,color='red')
    self.widget['gammacorr'].label.set_color('white')
    self.widget['gammacorr'].valtext.set_color('white')

    self.widget['robust'].on_changed(self._onRobust)
    self.widget['lat'].on_changed(self._onKeyLat)
    self.widget['dec'].on_changed(self._onKeyDec)
    self.widget['H0'].on_changed(self._onKeyH0)
    self.widget['H1'].on_changed(self._onKeyH1)
    self.widget['wave'].on_changed(self._changeWavelength)
    self.widget['add'].on_clicked(self._addAntenna)
    self.widget['rem'].on_clicked(self._removeAntenna)
    self.widget['save'].on_clicked(self.saveArray)
    self.widget['loadarr'].on_clicked(self.loadArray)
    self.widget['loadmod'].on_clicked(self.loadModel)
    self.widget['gammacorr'].on_changed(self._gammacorr)


    self._prepareBeam()
    self._prepareBaselines()
    self._setBaselines()
    self._setBeam()
    self._plotBeam()
    self._plotAntennas()
    self._prepareModel()
    self._plotModel()
    self._plotDirty()
    self._plotModelFFT()
    pl.show()



  def readAntennas(self,antenna_file):
    
    self.Hcov = [-12.0*self.Hfac,12.0*self.Hfac]
    self.Hmax = np.pi
    self.lat = 45.*self.deg2rad
    self.dec = 60.*self.deg2rad
    self.trlat = [np.sin(self.lat),np.cos(self.lat)]
    self.trdec = [np.sin(self.dec), np.cos(self.dec)]
    self.Xmax = 4.0

    if len(antenna_file)==0:
      self.Nant = 7
      self.antPos=[[0.0,0.0],[0.0,1.],[0.0,2.0],[1.,-1.],[2.0,-2.0],[-1.,-1.],[-2.0,-2.0]]
    

    if len(antenna_file)>0: 
     if not os.path.exists(antenna_file):
      self.showError("\n\nAntenna file %s does not exist!\n\n"%antenna_file)
      return False
    
     else:
      antPos=[]
      Hcov = [0,0]
      Nant = 0
      Xmax = 0.0
      fi = open(antenna_file)
      for li,l in enumerate(fi.readlines()):
        comm = l.find('#')
        if comm>=0:
          l = l[:comm]     
        it = l.split()
        if len(it)>0:
          if it[0]=='ANTENNA':   
            antPos.append(map(float,it[1:]))
            Nant += 1
            antPos[-1][0] *= 1.e-3 ; antPos[-1][1] *= 1.e-3
            Xmax = np.max(np.abs(antPos[-1]+[Xmax]))
          elif it[0]=='LATITUDE':
            lat = float(it[1])*self.deg2rad
            trlat = [np.sin(lat),np.cos(lat)]
          elif it[0]=='DECLINATION':
            dec = float(it[1])*self.deg2rad
            trdec = [np.sin(dec),np.cos(dec)]
          elif it[0]=='HOUR_ANGLE':
            Hcov[0] = float(it[1])*self.Hfac
            Hcov[1] = float(it[2])*self.Hfac
          else:
            self.showError("\n\nWRONG SYNTAX IN LINE %i:\n\n %s...\n\n"%(li+1,l[:max(10,len(l))]))


      if np.abs(lat-dec>=np.pi/2.):
         self.showError("\n\nSource is either not observable or just at the horizon!\n\n")
         return False
      if Nant<2:
         self.showError("\n\nThere should be at least 2 antennas!\n\n")
         return False



      self.Nant = Nant
      self.antPos = antPos
      self.lat = lat
      self.dec = dec
      self.trlat = trlat
      self.trdec = trdec
      self.Hcov = Hcov
      self.Xmax = Xmax


      cosW = -np.tan(self.lat)*np.tan(self.dec)
      if np.abs(cosW) < 1.0:
        Hhor = np.arccos(cosW)
      elif np.abs(self.lat-self.dec)>np.pi/2.:
        Hhor = 0
      else:
        Hhor = np.pi

      if Hhor>0.0:
        if self.Hcov[0]< -Hhor:
          self.Hcov[0] = -Hhor
        if self.Hcov[1]>  Hhor:
          self.Hcov[1] =  Hhor

      self.Hmax = Hhor
      H = np.linspace(self.Hcov[0],self.Hcov[1],self.nH)[np.newaxis,:]
      self.Xmax = Xmax*1.5
      fi.close()
   
    return True




  def readModels(self,model_file):

    self.wavelength = 1.e-6  # in km.
    self.imsize = 4. 
    self.imfiles = []

    if len(model_file)==0:
      self.models = [['G',0.,0.4,1.0,0.1],['D',0.,0.,2.,0.5],['P',-0.4,-0.5,0.1]]
      self.Xaxmax = self.imsize/2.
      return True

    if len(model_file)>0: 
     if not os.path.exists(model_file):
      self.showError("\n\nModel file %s does not exist!\n\n"%model_file)
      return False

     else:
      fixsize = False
      models = []
      imfiles = []
      Xmax = 0.0
      fi = open(model_file)
      for li,l in enumerate(fi.readlines()):
        comm = l.find('#')
        if comm>=0:
          l = l[:comm]     
        it = l.split()
        if len(it)>0:
          if it[0]=='IMAGE':
            imfiles.append([str(it[1]),float(it[2])])
          elif it[0] in ['G','D','P']:   
            models.append([it[0]]+map(float,it[1:]))
            if models[-1][0] != 'P':
              models[-1][4] = np.abs(models[-1][4])
              Xmax = np.max([np.abs(models[-1][1])+models[-1][4],
                 np.abs(models[-1][2])+models[-1][4],Xmax])
          elif it[0] == 'WAVELENGTH':
            wavelength = float(it[1])*1.e-3
          elif it[0] == 'IMSIZE':
            imsize = 2.*float(it[1])
            fixsize = True
          else:
            self.showError("\n\nWRONG SYNTAX IN LINE %i:\n\n %s...\n\n"%(li+1,l[:max(10,len(l))]))

      if len(models)+len(imfiles)==0:
         self.showError("\n\nThere should be at least 1 model component!\n\n")

      self.models=models
      self.wavelength=wavelength
      self.imsize=imsize
      self.imfiles = imfiles

      if not fixsize:
        self.imsize = Xmax*1.1


      self.Xaxmax = self.imsize/2.

      fi.close()



    return True





  def _changeWavelength(self,wave):

     self.wavelength = wave*1.e-6
     fmtB1 = r'$\lambda = $ %4.1fmm  '%(self.wavelength*1.e6)
     self.fmtB = fmtB1 + "\n" r'% 4.2f Jy/beam' "\n" r'$\Delta\alpha = $ % 4.2f / $\Delta\delta = $ % 4.2f '

     self._plotAntennas(redo=False)
     self._changeCoordinates(rescale=True)
     self._plotModelFFT(redo=False) 



  def _changeCoordinates(self,rescale=False):

    if self.lat>np.pi/2.:
      self.lat = np.pi/2.
      return
    elif self.lat<-np.pi/2.:
      self.lat = -np.pi/2.
      return

    if self.dec>np.pi/2.:
      self.dec = np.pi/2.
      return
    elif self.dec<-np.pi/2.:
      self.dec = -np.pi/2.
      return

    self.trlat = [np.sin(self.lat),np.cos(self.lat)]
    self.trdec = [np.sin(self.dec), np.cos(self.dec)]


    cosW = -np.tan(self.lat)*np.tan(self.dec)
    if np.abs(cosW) < 1.0:
      Hhor = np.arccos(cosW)
    elif np.abs(self.lat-self.dec)>np.pi/2.:
      Hhor = 0
    else:
      Hhor = np.pi

    self.Hmax = Hhor

    if self.Hmax>0.0:
      if self.Hcov[0]< -self.Hmax:
        self.Hcov[0] = -self.Hmax
        self.lock=True
        self.widget['H0'].set_val(self.Hcov[0]/self.Hfac)
        self.lock=False
      if self.Hcov[1]>  self.Hmax:
        self.Hcov[1] =  self.Hmax
        self.lock=True
        self.widget['H1'].set_val(self.Hcov[1]/self.Hfac)
        self.lock=False


    newtext = self.fmtH%(self.lat/self.deg2rad,self.dec/self.deg2rad,self.Hcov[0]/self.Hfac,self.Hcov[1]/self.Hfac)
    self.latText.set_text(newtext)
    self.Horig =  np.linspace(self.Hcov[0],self.Hcov[1],self.nH)
    H = self.Horig[np.newaxis,:]
    self.H = [np.sin(H),np.cos(H)] 
    self._setBaselines()
    self._setBeam()
    self._plotBeam(redo=False)
    self._plotAntennas(redo=False,rescale=True)
    self._plotDirty(redo=False)

    pl.draw()



  def _prepareBeam(self):
    
    self.beam = np.zeros((self.Npix,self.Npix),dtype=np.float32)
    self.totsampling = np.zeros((self.Npix,self.Npix),dtype=np.float32)
    self.dirtymap = np.zeros((self.Npix,self.Npix),dtype=np.float32)
    self.totsamplingU = np.zeros((self.Npix,self.Npix),dtype=np.float32)
    self.robustsamp = np.zeros((self.Npix,self.Npix),dtype=np.float32)


  def _prepareBaselines(self):

    self.Nbas = self.Nant*(self.Nant-1)/2
    NBmax = self.Nbas
    self.B = np.zeros((NBmax,self.nH),dtype=np.float32)
    self.basnum = np.zeros((self.Nant,self.Nant-1),dtype=np.int8)
    self.refant = np.zeros((self.Nant,self.Nant-1),dtype=np.int8)
    self.basidx = np.zeros((self.Nant,self.Nant),dtype=np.int8)
    self.antnum = np.zeros((NBmax,2),dtype=np.int8)

    self.Horig = np.linspace(self.Hcov[0],self.Hcov[1],self.nH)
    H = self.Horig[np.newaxis,:]
    self.H = [np.sin(H),np.cos(H)]

    bi = 0
    nii = [0 for n in range(self.Nant)]
    for n1 in range(self.Nant-1):
      for n2 in range(n1+1,self.Nant):
        self.basnum[n1,nii[n1]] = bi
        self.basnum[n2,nii[n2]] = bi
        self.basidx[n1,n2] = bi
        self.antnum[bi] = [n1,n2]
        self.refant[n1,nii[n1]] = 0
        self.refant[n2,nii[n2]] = 1
        nii[n1] += 1; nii[n2] += 1
        bi += 1

    self.u = np.zeros((NBmax,self.nH))
    self.v = np.zeros((NBmax,self.nH))
    self.ravelDims = (NBmax,self.nH)

    self.sampling = np.zeros((self.Nant,self.Npix,self.Npix),dtype=np.float32)



  def _setBaselines(self,antidx=-1):

   if antidx==-1:
     bas2change = range(self.Nbas)
   else:
     bas2change = self.basnum[antidx].flatten()

   for currBas in bas2change:
     n1,n2 = self.antnum[currBas]
     self.B[currBas,0] = -(self.antPos[n2][1]-self.antPos[n1][1])*self.trlat[0]/self.wavelength
     self.B[currBas,1] = (self.antPos[n2][0]-self.antPos[n1][0])/self.wavelength
     self.B[currBas,2] = (self.antPos[n2][1]-self.antPos[n1][1])*self.trlat[1]/self.wavelength
     self.u[currBas,:] = -(self.B[currBas,0]*self.H[0] + self.B[currBas,1]*self.H[1])
     self.v[currBas,:] = -self.B[currBas,0]*self.trdec[0]*self.H[1]+self.B[currBas,1]*self.trdec[0]*self.H[0]+self.trdec[1]*self.B[currBas,2]




  def _gridUV(self,antidx=-1):

   if antidx==-1:
     bas2change = range(self.Nbas)
     dummy = np.zeros(self.Npix*self.Npix,dtype=np.int8)
     self.pixU = [dummy for nb in bas2change]
     self.pixV = [dummy for nb in bas2change]
     self.totsampling[:] = 0.0
   else:
     bas2change = self.basnum[antidx].flatten()

  # Nphf = int(self.Npix/2)
   self.UVpixsize = 2./(self.imsize*np.pi/180./3600.)


   for nb in bas2change:
     pixU = np.rint(self.u[nb]/self.UVpixsize).flatten().astype(np.int32)
     pixV = np.rint(self.v[nb]/self.UVpixsize).flatten().astype(np.int32)
     goodpix = np.logical_and(np.abs(pixU)<self.Nphf,np.abs(pixV)<self.Nphf)
     if not antidx==-1:
      for pi in range(len(self.pixU[nb])):
       self.totsampling[self.pixV[nb][pi]+self.Nphf,-self.pixU[nb][pi]+self.Nphf] -= 1.0
       self.totsampling[-self.pixV[nb][pi]+self.Nphf,self.pixU[nb][pi]+self.Nphf] -= 1.0

     self.pixU[nb] = np.copy(pixU[goodpix])
     self.pixV[nb] = np.copy(pixV[goodpix])
     for pi in range(len(self.pixU[nb])):
       self.totsampling[self.pixV[nb][pi]+self.Nphf,-self.pixU[nb][pi]+self.Nphf] += 1.0
       self.totsampling[-self.pixV[nb][pi]+self.Nphf,self.pixU[nb][pi]+self.Nphf] += 1.0

   self.totsamplingU[:] = 0.0
   self.totsamplingU[self.totsampling>0.0] = 1.0
   self.robfac = (5.*10.**(-self.robust))**2.*(2.*self.Nbas*self.nH)/np.sum(self.totsampling**2.)


  def _setBeam(self,antidx=-1):

  # print 'A:',self.robfac
   self._gridUV(antidx=antidx) 
  # print 'B:',self.robfac

   self.robustsamp[:] = self.totsampling/(1.+self.robfac*self.totsampling)


   self.beam[:] = np.fft.ifftshift(np.fft.ifft2(np.fft.fftshift(self.robustsamp))).real 
  # Nphf = int(self.Npix/2)
   self.beamScale = np.max(self.beam[self.Nphf:self.Nphf+1,self.Nphf:self.Nphf+1])
   self.beam[:] /= self.beamScale

  # print 'C:',np.std(self.beam)


  def _prepareModel(self):

    pixsize = float(self.imsize)/self.Npix
    xx = np.linspace(-self.imsize/2.,self.imsize/2.,self.Npix)
    yy = np.ones(self.Npix,dtype=np.float32)
    distmat = np.zeros((self.Npix,self.Npix),dtype=np.float32)
    self.modelim = np.zeros((self.Npix,self.Npix),dtype=np.float32)
   # Nphf = int(self.Npix/2)
    for model in self.models:
      xsh = -model[1]
      ysh = -model[2]
      xpix = np.rint(xsh/pixsize).astype(np.int32)
      ypix = np.rint(ysh/pixsize).astype(np.int32)
      centy = np.roll(xx,ypix)
      centx = np.roll(xx,xpix)
      distmat[:] = np.outer(centy**2.,yy) + np.outer(yy,centx**2.)
      if model[0]=='D':
        mask = np.logical_or(distmat<=model[4]**2.,distmat==np.min(distmat))
        self.modelim[mask] += float(model[3])/np.sum(mask)
      elif model[0]=='G':
        gauss = np.exp(-distmat/(2.*model[4]**2.))
        self.modelim[:] += float(model[3])*gauss/np.sum(gauss)
      elif model[0]=='P':
        if np.abs(xpix+self.Nphf)<self.Npix and np.abs(ypix+self.Nphf)<self.Npix:
          yint = ypix+self.Nphf
          xint = xpix+self.Nphf
       #   print ypix+Nphf, xpix+Nphf
          self.modelim[yint,xint] += float(model[3])

    for imfile in self.imfiles:
      if not os.path.exists(imfile[0]):
       # showinfo('File missing!','File %s does NOT exist. \nWill try to look in the %s directory'%(imfile[0],self.datadir)) 
        imfile[0] = os.path.join(self.datadir,imfile[0])
        if not os.path.exists(imfile[0]):
          self.showError('File %s does NOT exist. Cannot read the model!'%imfile[0]) 
          return

      Np4 = self.Npix/4
      img = plimg.imread(imfile[0]).astype(np.float32)
      dims = np.shape(img)
      d3 = min(2,dims[2])
      d1 = float(max(dims))
      avimg = np.average(img[:,:,:d3],axis=2)
      avimg -= np.min(avimg)
      avimg *= imfile[1]/np.max(avimg)
      if d1 == self.Nphf:
        sh0 = (self.Nphf-dims[0])/2
        sh1 = (self.Nphf-dims[1])/2
        self.modelim[sh0+Np4:sh0+Np4+dims[0], sh1+Np4:sh1+Np4+dims[1]] += zoomimg
      else:
        zoomimg = spndint.zoom(avimg,float(self.Nphf)/d1)
        zdims = np.shape(zoomimg)
        zd0 = min(zdims[0],self.Nphf)
        zd1 = min(zdims[1],self.Nphf)
    #    print zd0, zd1, zdims, dims
        sh0 = (self.Nphf-zdims[0])/2
        sh1 = (self.Nphf-zdims[1])/2
        self.modelim[sh0+Np4:sh0+Np4+zd0, sh1+Np4:sh1+Np4+zd1] += zoomimg[:zd0,:zd1]

    #  figg = pl.figure()
    #  pl.imshow(self.modelim[Np4:self.Npix-Np4,Np4:self.Npix-Np4])
    #  pl.show()

    self.modelim[self.modelim<0.0] = 0.0
    self.modelfft = np.fft.fft2(np.fft.fftshift(self.modelim))


  def _plotModel(self,redo=True):
    
    Np4 = self.Npix/4

    if redo:
      self.modelPlotPlot = self.modelPlot.imshow(np.power(self.modelim[Np4:self.Npix-Np4,Np4:self.Npix-Np4],self.gamma),picker=True,interpolation='nearest',vmin=0.0,vmax=np.max(self.modelim)**self.gamma,cmap=cm.jet)
 #     self.modelPlotPlot = self.modelPlot.imshow(self.modelim[Np4:self.Npix-Np4,Np4:self.Npix-Np4],picker=True,interpolation='nearest') #,vmin=0.0,vmax=np.max(self.modelim))

      modflux = self.modelim[self.Nphf,self.Nphf]
      self.modelText = self.modelPlot.text(0.05,0.87,self.fmtM%(modflux,0.0,0.0),
         transform=self.modelPlot.transAxes,bbox=dict(facecolor='white', 
         alpha=0.7))
      pl.setp(self.modelPlotPlot, extent=(self.Xaxmax/2.,-self.Xaxmax/2.,-self.Xaxmax/2.,self.Xaxmax/2.))
      self.modelPlot.set_ylabel('Dec offset (as)')
      self.modelPlot.set_xlabel('RA offset (as)')
      self.modelPlot.set_title('MODEL IMAGE')
    else:
      self.modelPlotPlot.set_data(np.power(self.modelim[Np4:self.Npix-Np4,Np4:self.Npix-Np4],self.gamma))
  #    self.modelPlotPlot.set_data(self.modelim[Np4:self.Npix-Np4,Np4:self.Npix-Np4])
      extr = [0.0,np.max(self.modelim)**self.gamma]
      self.modelPlotPlot.norm.vmin = extr[0]
      self.modelPlotPlot.norm.vmax = extr[1]
      pl.setp(self.modelPlotPlot, extent=(self.Xaxmax/2.,-self.Xaxmax/2.,-self.Xaxmax/2.,self.Xaxmax/2.))



  def _plotModelFFT(self,redo=True):

    self.UVmax = self.Npix/2./self.lfac*self.UVpixsize
    self.FFTtoplot = np.fft.fftshift(self.modelfft)
    toplot = np.abs(self.FFTtoplot)
    mval = np.min(toplot)
    Mval = np.max(toplot)
    dval = (Mval-mval)/2.
 #   print 'MVALS: ',mval,Mval

    if redo:
       mymap = pl.gray()
       self.UVPlotFFTPlot = self.UVPlot.imshow(toplot,cmap=mymap,vmin=mval-dval,vmax=Mval+dval,picker=5)
       pl.setp(self.UVPlotFFTPlot, extent=(-self.UVmax/2.,self.UVmax/2.,-self.UVmax/2.,self.UVmax/2.))
    else:
       self.UVPlotFFTPlot.set_data(toplot)
    #   pl.setp(self.UVPlotFFTPlot, extent=(-self.UVmax/2.,self.UVmax/2.,-self.UVmax/2.,self.UVmax/2.))
       self.UVPlotFFTPlot.norm.vmin = mval-dval
       self.UVPlotFFTPlot.norm.vmax = Mval+dval

    


  def _plotDirty(self,redo=True):
    Np4 = self.Npix/4

    self.dirtymap[:] = np.fft.fftshift(np.fft.ifft2(self.modelfft*np.fft.ifftshift(self.robustsamp))).real/self.beamScale
    extr = [np.min(self.dirtymap),np.max(self.dirtymap)]
   # Nphf = int(self.Npix/2)
    if redo:
      self.dirtyPlotPlot = self.dirtyPlot.imshow(self.dirtymap[Np4:self.Npix-Np4,Np4:self.Npix-Np4],interpolation='nearest',picker=True)
      modflux = self.dirtymap[self.Nphf,self.Nphf]
      self.dirtyText = self.dirtyPlot.text(0.05,0.87,self.fmtD%(modflux,0.0,0.0),
         transform=self.dirtyPlot.transAxes,bbox=dict(facecolor='white', 
         alpha=0.7))
      pl.setp(self.dirtyPlotPlot, extent=(self.Xaxmax/2.,-self.Xaxmax/2.,-self.Xaxmax/2.,self.Xaxmax/2.))
      self.dirtyPlot.set_ylabel('Dec offset (as)')
      self.dirtyPlot.set_xlabel('RA offset (as)')
      self.dirtyPlot.set_title('DIRTY IMAGE')
    else:
      self.dirtyPlotPlot.set_data(self.dirtymap[Np4:self.Npix-Np4,Np4:self.Npix-Np4])
      self.dirtyPlotPlot.norm.vmin = extr[0]
      self.dirtyPlotPlot.norm.vmax = extr[1]
  #    pl.setp(self.dirtyPlotPlot, extent=(self.Xaxmax/2.,-self.Xaxmax/2.,-self.Xaxmax/2.,self.Xaxmax/2.))






  def _plotAntennas(self,redo=True,rescale=False):

   mw = 2.*self.Xmax/self.wavelength/self.lfac
   if mw < 0.1 and self.lfac == 1.e6:
      self.lfac = 1.e3
    #  self.Xmax *= 1.e-3
      self.ulab = r'U (k$\lambda$)'
      self.vlab = r'V (k$\lambda$)'
   elif mw >=100. and self.lfac == 1.e3:
      self.lfac = 1.e6
    #  self.Xmax *= 1.e3
      self.ulab = r'U (M$\lambda$)'
      self.vlab = r'V (M$\lambda$)'
   

   if redo:
    toplot = np.array(self.antPos[:self.Nant])
    self.antPlotPlot = self.antPlot.plot(toplot[:,0],toplot[:,1],'og', picker=5)[0]
    self.antPlot.set_xlim((-self.Xmax,self.Xmax))
    self.antPlot.set_ylim((-self.Xmax,self.Xmax))
    self.antPlot.set_xlabel('E-W offset (km)')
    self.antPlot.set_ylabel('N-S offset (km)')
    self.antPlot.set_title('ARRAY CONFIGURATION')
    self.antText = self.antPlot.text(0.05,0.92,self.fmtA%self.Nant,transform=self.antPlot.transAxes)
    self.UVPlotPlot = []
    toplotu = self.u.flatten()/self.lfac ;  toplotv = self.v.flatten()/self.lfac ; 
    self.UVPlotPlot.append(self.UVPlot.plot(toplotu, toplotv,'.b',markersize=1,picker=2)[0])
    self.UVPlotPlot.append(self.UVPlot.plot(-toplotu,-toplotv,'.r',markersize=1,picker=2)[0])
    self.UVPlot.set_xlim((2.*self.Xmax/self.wavelength/self.lfac,-2.*self.Xmax/self.wavelength/self.lfac))
    self.UVPlot.set_ylim((2.*self.Xmax/self.wavelength/self.lfac,-2.*self.Xmax/self.wavelength/self.lfac))
    self.latText = self.UVPlot.text(0.05,0.87,self.fmtH%(self.lat/self.deg2rad,self.dec/self.deg2rad,self.Hcov[0]/self.Hfac,self.Hcov[1]/self.Hfac),transform=self.UVPlot.transAxes)
    self.latText.set_color('orange')
    self.basText = self.UVPlot.text(0.05,0.02,self.fmtBas%(0,0,0.0),transform=self.UVPlot.transAxes)
    self.visText = self.UVPlot.text(0.05,0.08,self.fmtVis%(0.0,0.0),transform=self.UVPlot.transAxes)
    self.visText.set_color('orange')

    self.basText.set_color('orange')
    self.UVPlot.set_xlabel(self.ulab)
    self.UVPlot.set_ylabel(self.vlab) 
    self.UVPlot.set_title('UV PLANE')

    self.antLabelPlot = []

    for i in range(self.Nant):
      self.antLabelPlot.append(self.antPlot.annotate(str(i+1),textcoords = 'offset points',xy=(toplot[i,0],toplot[i,1]),xytext=(-7,4)))
  

   else:

    toplot = np.array(self.antPos[:self.Nant])
    self.antPlotPlot.set_data(toplot[:,0],toplot[:,1])
    toplotu = self.u.flatten()/self.lfac ;  toplotv = self.v.flatten()/self.lfac ; 
    if rescale:
      self.antPlot.set_xlim((-self.Xmax,self.Xmax))
      self.antPlot.set_ylim((-self.Xmax,self.Xmax))
      self.UVPlot.set_xlim((2.*self.Xmax/self.wavelength/self.lfac,-2.*self.Xmax/self.wavelength/self.lfac))
      self.UVPlot.set_ylim((2.*self.Xmax/self.wavelength/self.lfac,-2.*self.Xmax/self.wavelength/self.lfac))


    if len(self.antLabelPlot)>self.Nant:
      for i in range(self.Nant,len(self.antLabelPlot)):
         self.antLabelPlot[i].set_visible(False)


    for i in range(self.Nant):
      if i>len(self.antLabelPlot)-1:
        self.antLabelPlot.append(self.antPlot.annotate(str(i+1),textcoords = 'offset points',xy=(toplot[i,0],toplot[i,1]),xytext=(-7,4)))
      else:
        self.antLabelPlot[i].set_visible(True)
        self.antLabelPlot[i].xy = (toplot[i,0],toplot[i,1])
    

    self.UVPlotPlot[0].set_data(toplotu,toplotv)
    self.UVPlotPlot[1].set_data(-toplotu,-toplotv)
    self.UVPlot.set_xlabel(self.ulab)
    self.UVPlot.set_ylabel(self.vlab) 

    

  def _plotBeam(self,redo=True):

    Np4 = self.Npix/4
    if redo:
      self.beamPlotPlot = self.beamPlot.imshow(self.beam[Np4:self.Npix-Np4,Np4:self.Npix-Np4],picker=True,interpolation='nearest')
      self.beamText = self.beamPlot.text(0.05,0.80,self.fmtB%(1.0,0.0,0.0),
           transform=self.beamPlot.transAxes,bbox=dict(facecolor='white', alpha=0.7))
      self.beamPlot.set_ylabel('Dec offset (as)')
      self.beamPlot.set_xlabel('RA offset (as)')
      pl.setp(self.beamPlotPlot, extent=(self.Xaxmax/2.,-self.Xaxmax/2.,-self.Xaxmax/2.,self.Xaxmax/2.))
      self.beamPlot.set_title('DIRTY BEAM')
      pl.draw()
    else:
      self.beamPlotPlot.set_data(self.beam[Np4:self.Npix-Np4,Np4:self.Npix-Np4])
      self.beamText.set_text(self.fmtB%(1.0,0.0,0.0))
   #   pl.setp(self.beamPlotPlot, extent=(self.Xaxmax/2.,-self.Xaxmax/2.,-self.Xaxmax/2.,self.Xaxmax/2.))


    self.nptot = np.sum(self.totsamplingU[:])
    self.beamPlotPlot.norm.vmin = np.min(self.beam)
    self.beamPlotPlot.norm.vmax = 1.0

    if np.sum(self.totsamplingU[self.Nphf-8:self.Nphf+8,self.Nphf-8:self.Nphf+8])==self.nptot:
      warn = 'WARNING!\nToo short baselines for such a small image\nPLEASE, INCREASE THE IMAGE SIZE!\nAND/OR DECREASE THE WAVELENGTH'
      self.beamText.set_text(warn)

    self.spherePlot.view_init(elev=self.dec/self.deg2rad,azim=0)
    self.arrayPath[0][:] = 10.*self.H[1]*np.cos(self.lat)
    self.arrayPath[1][:] = 10.*self.H[0]*np.cos(self.lat)
    self.arrayPath[2][:] = 10.*np.sin(self.lat)
    self.sphereArray[0].set_data(self.arrayPath[0], self.arrayPath[1])
    self.sphereArray[0].set_3d_properties(self.arrayPath[2])



  def _onPick(self,event):

   if event.mouseevent.inaxes == self.UVPlot:

     if event.artist in self.UVPlotPlot:
       idata = np.unravel_index(event.ind,self.ravelDims)
       if event.artist == self.UVPlotPlot[0]:
          n1,n2 = self.antnum[idata[0][0]]
       else:
          n2,n1 = self.antnum[idata[0][0]]

       H = self.Horig[idata[1][0]]/self.Hfac

       newtext = self.fmtBas%(n1+1,n2+1,H)
       self.basText.set_text(newtext)
       pl.draw()
     else:
       Up = event.mouseevent.xdata
       Vp = event.mouseevent.ydata
       yi = np.floor((self.UVmax/2.+Up)/(self.UVmax)*self.Npix)
       xi = np.floor((self.UVmax/2.-Vp)/(self.UVmax)*self.Npix)
       Flux = self.FFTtoplot[xi,yi]
       Phas, Amp = np.angle(Flux,deg=True), np.abs(Flux)
       newtext = self.fmtVis%(Amp,Phas)
       self.visText.set_text(newtext)
       pl.draw()
   #    print Up,Vp
#       self.auxVar = event

   elif event.mouseevent.inaxes == self.beamPlot:

     RA = event.mouseevent.xdata
     Dec = event.mouseevent.ydata
     yi = np.floor((self.Xaxmax-RA)/(2.*self.Xaxmax)*self.Npix)
     xi = np.floor((self.Xaxmax-Dec)/(2.*self.Xaxmax)*self.Npix)
     Flux = self.beam[xi,yi]
     self.beamText.set_text(self.fmtB%(Flux,RA,Dec))
     pl.draw()

   elif event.mouseevent.inaxes == self.dirtyPlot:

     RA = event.mouseevent.xdata
     Dec = event.mouseevent.ydata
     yi = np.floor((self.Xaxmax-RA)/(2.*self.Xaxmax)*self.Npix)
     xi = np.floor((self.Xaxmax-Dec)/(2.*self.Xaxmax)*self.Npix)
     Flux = self.dirtymap[xi,yi]
     self.dirtyText.set_text(self.fmtD%(Flux,RA,Dec))
     pl.draw()

   elif event.mouseevent.inaxes == self.modelPlot:

     RA = event.mouseevent.xdata
     Dec = event.mouseevent.ydata
     yi = np.floor((self.Xaxmax-RA)/(2.*self.Xaxmax)*self.Npix)
     xi = np.floor((self.Xaxmax-Dec)/(2.*self.Xaxmax)*self.Npix)
     Flux = self.modelim[xi,yi]
     self.modelText.set_text(self.fmtM%(Flux,RA,Dec))
     pl.draw()

   elif event.mouseevent.inaxes == self.antPlot:

    if event.mouseevent.button==1 and not self.pickAnt:
     self.antidx = event.ind
     if len(self.antidx) > 1:
       self.antidx = self.antidx[-1]
     self.pickAnt = True
     self.antText.set_text(self.fmtA%self.Nant + self.fmtA2%self.antidx)
     pl.draw()




  def _onAntennaDrag(self,event):   
     if self.pickAnt:
       self.antPos[self.antidx] = [event.xdata,event.ydata]
       self._setBaselines(antidx=self.antidx)
       self._plotAntennas(redo=False)
       self._setBeam(antidx=self.antidx)
       self._plotBeam(redo=False)
       self._plotDirty(redo=False)
       pl.draw()


# Drag the sphere plot (to change source position)
     if self._onSphere:
       oldDec = self.dec/self.deg2rad
       newDec, newH0 = self.spherePlot.elev, self.spherePlot.azim

   # Limits on declination:
       if np.abs(newDec)>90.:
         self.spherePlot.view_init(elev = oldDec, azim=0.0)
         return

       newDec *= self.deg2rad
       if not self.lock:
         self.lock=True
         if newDec != self.dec and np.abs(newDec-self.lat)<np.pi/2.:
           self.widget['dec'].set_val(newDec/self.deg2rad)
           self.spherePlot.view_init(elev = newDec/self.deg2rad, azim=0.0)
           self.dec = newDec
           self._changeCoordinates()
           self.lock = False
         else:
           self.spherePlot.view_init(elev = oldDec, azim=0.0)
           self.lock = False
       else:
           self.spherePlot.view_init(elev = oldDec, azim=0.0)





  def _onRelease(self,event):
     self._onSphere = False
     if self.pickAnt:
       self.pickAnt = False
       self.antText.set_text(self.fmtA%self.Nant)
       pl.draw()


  def _onRobust(self,newrob):
    self.robust = newrob
    self._changeCoordinates()


  def _onKeyLat(self,newlat):
   newlat *= self.deg2rad
   if not self.lock:
    self.lock = True
    if newlat != self.lat: 
     if np.abs(newlat-self.dec)<np.pi/2.:
      self.lat = newlat
      self._changeCoordinates()
     else:
      self.widget['lat'].set_val(self.lat/self.deg2rad)
    self.lock = False


  def _gammacorr(self,gamma):
    self.gamma = gamma
    self._plotModel(redo=False)
    pl.draw()

  def _onKeyDec(self,newdec):
   newdec *= self.deg2rad
   if not self.lock:
    self.lock=True
    if newdec != self.dec: 
     if np.abs(newdec-self.lat)<np.pi/2.:
      self.dec = newdec
      self._changeCoordinates()
     else:
      self.widget['dec'].set_val(self.dec/self.deg2rad)
    self.lock = False



  def _onKeyH0(self,newH0):
   newH0 *= self.Hfac
   if not self.lock:
    self.lock = True
    if np.abs(newH0) < self.Hmax:
      self.Hcov[0] = newH0
    else:
      self.Hcov[0] = -self.Hmax
      self.widget['H0'].set_val(self.Hcov[0]/self.Hfac) 
    if self.Hcov[1]<self.Hcov[0]:
      self.Hcov[1] = self.Hcov[0]
      self.widget['H1'].set_val(self.Hcov[1]/self.Hfac) 
    self._changeCoordinates()
    self.lock=False



  def _onKeyH1(self,newH1):
   newH1 *= self.Hfac
   if not self.lock:
    self.lock = True
    if np.abs(newH1) < self.Hmax:
      self.Hcov[1] = newH1
    else:
      self.Hcov[1] = self.Hmax
      self.widget['H1'].set_val(self.Hcov[1]/self.Hfac) 
    if self.Hcov[0]>self.Hcov[1]:
      self.Hcov[0] = self.Hcov[1]
      self.widget['H0'].set_val(self.Hcov[0]/self.Hfac) 
    self._changeCoordinates()
    self.lock=False


  def _addAntenna(self,antenna):

    if self.Nant >= len(self.antPos):
      self.antPos.append([0.,0.])
      self.Nant += 1
      self.antLabelPlot.append(self.antPlot.annotate(str(self.Nant),textcoords = 'offset points',xy=(0,0),xytext=(-7,4)))

    else:
      self.antLabelPlot[self.Nant].xy = (self.antPos[self.Nant][0],self.antPos[self.Nant][1])
      self.Nant += 1



    self.antLabelPlot[self.Nant-1].set_visible(True)

    newtext = self.fmtA%self.Nant 
    self.antText.set_text(newtext)
    self._prepareBaselines()
    self._changeCoordinates()


  def _removeAntenna(self,antenna):
    if self.Nant > 2:
        self.Nant -= 1
        for i in range(self.Nant,len(self.antLabelPlot)):
           self.antLabelPlot[i].set_visible(False)
    newtext = self.fmtA%self.Nant 
    self.antText.set_text(newtext)
    self._prepareBaselines()
    self._changeCoordinates()


  def _onPress(self,event):
    if event.inaxes == self.spherePlot:
      self._onSphere = True


  def saveArray(self,array):

    fname = tkFileDialog.asksaveasfilename(defaultextension='.array',title='Save current array...')
    iff = open(fname,'w')

    print >> iff,'LATITUDE % 3.1f'%(self.lat/self.deg2rad)
    print >> iff,'DECLINATION % 3.1f'%(self.dec/self.deg2rad)
    toprint = tuple([l/self.Hfac for l in self.Hcov])
    print >> iff,'HOUR_ANGLE % 3.1f  % 3.1f'%toprint
    for ant in self.antPos:
      toprint = tuple([p*1.e3 for p in ant])
      print >> iff,'ANTENNA  % .3e   % .3e'%toprint

    self.antText.set_text('SAVED: %s'%os.path.basename(fname))
    pl.draw()
    time.sleep(3)
    self.antText.set_text(self.fmtA%self.Nant) 
    pl.draw()
    iff.close()

          
    return


  def loadArray(self,array):

    antenna_file = tkFileDialog.askopenfilename(title='Load array...',initialdir=self.arraydir)
    self.lock=False

    if len(antenna_file)>0:
      goodread = self.readAntennas(str(antenna_file))

      if goodread:
        self._prepareBaselines()
        self._changeCoordinates()
        self._plotModelFFT(redo=False) 
        newtext = self.fmtA%self.Nant 
        self.antText.set_text(newtext)
        self.widget['lat'].set_val(self.lat/self.deg2rad)
        self.widget['dec'].set_val(self.dec/self.deg2rad)
        self.widget['H0'].set_val(self.Hcov[0]/self.Hfac)
        self.widget['H1'].set_val(self.Hcov[1]/self.Hfac)
        self.widget['wave'].set_val(self.wavelength*1.e6)
        self._plotAntennas(redo=False,rescale=True)
        pl.draw()

    return


  def loadModel(self,model):

    model_file = tkFileDialog.askopenfilename(title='Load model...',initialdir=self.modeldir)
    self.lock=False

    if len(model_file)>0:
      goodread = self.readModels(str(model_file))

      if goodread:
        self._prepareModel()
        self._plotModel(redo=False)
        self.widget['wave'].set_val(self.wavelength*1.e6)
        self._changeCoordinates()
        self._plotModelFFT(redo=False) 
        self.wax['wave'].cla()
        self.widget['wave'] = Slider(self.wax['wave'],r'$\lambda$ (mm)',self.wavelength*1.e6/10.,10.*self.wavelength*1.e6,valinit=self.wavelength*1.e6)
        self.widget['wave'].on_changed(self._changeWavelength)
        pl.draw()


    return





if __name__ == "__main__":
  myint = Interferometer()

